/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Test pthread signal support, with tests for cleanup handlers and
 * pthread cancelation thrown in.
 */

/*
 * Set these accordingly. Also, make sure main() inits the appropriate
 * drivers.
 */
#define DISKNAME "wd1"
#define PARTITION "b"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <string.h>
#include <oskit/threads/pthread.h>
#include <oskit/startup.h>
#include <assert.h>

#ifdef  OSKIT_UNIX
void	start_fs_native_pthreads(char *root);
void	start_network_native_pthreads(void);
#endif

void	sigtest(void);

void
main()
{
	/*
	 * Initialize the pthreads module.
	 */
	pthread_init(1);
	start_clock();

#ifndef OSKIT_UNIX
	osenv_process_lock();
	oskit_linux_init_ethernet_tulip();
	oskit_linux_init_ethernet_eepro100();
	oskit_linux_init_ide();
	oskit_dev_probe();
	osenv_process_unlock();
	start_fs_pthreads(DISKNAME, PARTITION);
	start_network_pthreads();
#else
	/*
	 * Unix mode.
	 */
	start_fs_native_pthreads("/");
	start_network_native_pthreads();
#endif
	sigtest();
}

/*
 * Test pthread signal support. Create some threads, have them install
 * signal handlers, then send it some signals.
 */
#define NUMTESTERS	5
#define NUMCYCLES	5
volatile int		alarmed;

void
sighandler(int sig, int code, struct sigcontext *scp)
{
	printf("SigHandler: pthread:%d sig:%d code:%d scp:%p\n",
	       pthread_self(), sig, code, scp);
}

void
cleaner(void *arg)
{
	printf("Cleaner: pthread:%d arg:%d\n", pthread_self(), (int) arg);
}

void
sigalarmer(int sig, int code, struct sigcontext *scp)
{
	printf("SigAlarmer: pthread:%d sig:%d code:%d scp:%p\n",
	       pthread_self(), sig, code, scp);
	alarmed = 1;
}

/*
 * The thread body for the alarm() tests.
 */
void
alarmtester(void *arg)
{
	int		i;

	for (i = 0; i < 5; i++) {
		printf("alarmtester:%d Calling Alarm ...\n", pthread_self());
		alarmed = 0;
		alarm(1);
		while (!alarmed)
			;
	}
	
	pthread_exit(0);
}

/*
 * The thread body for the pthread_kill tests.
 */
void
sigtester(void *arg)
{
	pthread_cleanup_push(cleaner, arg);

	printf("sigtester:%d ready\n", pthread_self());
	oskit_pthread_sleep(100);
	
	while (1) {
		oskit_pthread_sleep(20);
	}

	pthread_cleanup_pop(1);
}

void
sigtest(void)
{
	int		i, j;
	void		*stat;
	pthread_t	tids[NUMTESTERS];
	struct sigaction sa;

	/*
	 * Install SIGUSR1 handler for pthread_kill tests.
	 */
	sa.sa_handler = sighandler;
	sa.sa_flags   = 0;
	sa.sa_mask    = 0;

	if (sigaction(SIGUSR1, &sa, 0) < 0) {
		perror("sigaction SIGUSR1");
		exit(1);
	}

	/*
	 * Install SIGALRM handler for alarm() tests.
	 */
	sa.sa_handler = sigalarmer;
	sa.sa_flags   = 0;
	sa.sa_mask    = 0;

	if (sigaction(SIGALRM, &sa, 0) < 0) {
		perror("sigaction SIGALRM");
		exit(1);
	}

	/*
	 * Create a thread for the alarm test. Wait for it to do its
	 * thing, and then go onto the pthread_kill test.
	 */
	pthread_create(&tids[0], 0, alarmtester, (void *) 0);
	pthread_join(tids[0], &stat);

	/*
	 * Create threads for pthread_kill tests.
	 */
	for (i = 0; i < NUMTESTERS; i++) {
		pthread_create(&tids[i], 0, sigtester, (void *) i);
		oskit_pthread_sleep(20);
	}

	/*
	 * Send some signals
	 */
	for (j = 0; j < NUMCYCLES; j++) 
		for (i = 0; i < NUMTESTERS; i++) {
			pthread_kill(tids[i], SIGUSR1);
			oskit_pthread_sleep(20);
		}

	/*
	 * Cancel the threads.
	 */
	for (i = 0; i < NUMTESTERS; i++) {
		pthread_cancel(tids[i]);
	}

	for (i = 0; i < NUMTESTERS; i++) {
		pthread_join(tids[i], &stat);
	}

	oskit_pthread_sleep((long long) 100);
	printf("exiting ...\n");
	exit(0);
	return;
}
