/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Internal definitions for threads package.
 */
#ifndef	_PTHREAD_INT_H_
#define _PTHREAD_INT_H_

#ifndef ASSEMBLER
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>

/*
 * Turn on deadlock detection in spinlocks and interrupt mask checks.
 * Also turn on stackguards and deadlock detection.
 */
#ifdef  DEBUG
#undef  THREADS_DEBUG
#define THREADS_DEBUG
#endif

#ifdef	USERMODE
#define OSKIT_INLINE	static inline
#endif

#include <oskit/compiler.h>
#include <oskit/base_critical.h>

#ifdef hp800
#include <threads/hp800/pcb.h>
#else
#include <threads/x86/pcb.h>
#endif

#include <oskit/queue.h>
#include <oskit/debug.h>
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/clock.h>
#include <oskit/dev/timer.h>
#include <oskit/threads/pthread.h>
#include <threads/pthread_locking.h>
#ifdef  CPU_INHERIT
#include <threads/cpuinherit/pthread_cpuinherit.h>
#endif

#ifdef	SMP
#include <oskit/smp.h>
#endif

/*
 * Thread structure. The PCB is stored in the thread structure, but
 * accessed indirectly to leave room for future expansion.
 */
typedef struct pthread_thread {
	queue_chain_t		runq;		/* Runq/free link */
	pcb_t			*ppcb;		/* Pointer to PCB */
	oskit_u32_t		*pstk;		/* Pointer to stack alloc */
	void			*cookie;	/* Argument to thread */
	size_t			ssize;		/* Allocated stack size */
	size_t			guardsize;	/* Stack Guard size */
	pthread_t		tid;		/* Thread ID */
	int			interrupt_state;/* Interrupts and Preemption */
	oskit_u32_t		flags;		/* Thread state flags */
	pthread_lock_t		lock;		/* Lock */
	queue_chain_t		chain;		/* Queuing element */
	int			preempt;	/* Preempt this thread */

	/*
	 * The exit code is painful.
	 */
	oskit_u32_t		exitval;	/* Exit value for joiner */
	struct pthread_thread  *joining_me;	/* Back pointer to waiter */
	struct pthread_thread  *joining_with;	/* Target of the join */

	/*
	 * Resource accounting stuff. Added for the demo!
	 */
	int			cputime;	/* Total ticks consumed */
	int			cpticks;	/* Ticks in last second */
	oskit_u32_t		pctcpu;		/* Percent of CPU usage */
	int			childtime;	/* Accumulation of child CPU */

	/*
	 * To implement cancelation, must know the object blocked on.
	 */
	pthread_cond_t		*waitcond;	/* Condition variable */

	/*
	 * IPC stuff.
	 */
	struct {
		void		        *msg;		/* Message pointer */
		oskit_size_t		msg_size;
		pthread_t		tid;
		void			*reply;
		oskit_size_t		reply_size;
		queue_head_t		senders;
		queue_chain_t		senders_chain;
	} ipc_state;

	/*
	 * Timed condition wait is implemented using a clock timer. 
	 */
	struct oskit_timer	*condtimer;	

	/*
	 * pthread sleep/wakeup is implemented using a separate clock timer.
	 */
	struct oskit_timer	*sleeptimer;

	/*
	 * The osenv_sleep sleep record. Stashed to facilitate interrupt.
	 */
	struct osenv_sleeprec	*sleeprec;	/* pthread in an osenv_sleep */

	/*
	 * Signal stuff.
	 */
	sigset_t		sigmask;	/* Blocked signals */
	sigset_t		sigpending;	/* Pending signals */
	oskit_u32_t		eip;		/* For page fault handling */

	/*
	 * Keys values are currently a fixed array.
	 */
	void			*keyvalues[PTHREAD_KEYS_MAX]; /* Key/Value */

	/*
	 * Cancelation state and cleanup handlers.
	 */
	pthread_cleanup_t	*cleanups;	/* Cleanup handlers chain */
	char			cancelstate;	/* Cancelation State */
	char			canceltype;	/* Cancelation Type */
	
#ifdef  DEFAULT_SCHEDULER
	int			policy;		/* Scheduling Policy */
	int			priority;	/* Current Priority */
	int			base_priority;  /* Original priority */
	int			ticks;		/* Scheduling ticks left */
#endif
#ifdef  PRI_INHERIT
	/*
	 * Priority inheritance support.
	 */
	queue_head_t		waiters;	/* Threads waiting on me */
	queue_chain_t		waiters_chain;	/* Waiters queueing element */
	struct pthread_thread	*waiting_for;	/* Thread being waited for */
	struct pthread_thread	*inherits_from;	/* Thread inheriting from */
#endif
#ifdef  CPU_INHERIT
	/*
	 * CPU inheritance support.
	 */
	struct pthread_thread   *scheduler;	/* The threads scheduler */
	schedmsg_t		unblockmsg;
	schedmsg_queue_t	*msgqueue;	
	sched_wakecond_t	wakeup_cond;	/* Wakeup condition */
	schedflags_t		schedflags;	/* Flags */
	int			donate_rc;	/* Return code from donation */
	int			timeout;	/* In milliseconds */
	queue_head_t		donors;		/* Threads donating to me */
	queue_chain_t		donors_chain;	/* Donors queueing element */
	struct pthread_thread	*donating_to;	/* Thread being donated to */
	struct pthread_thread	*inherits_from;	/* Thread inheriting from */
	struct pthread_thread	*nextup;	/* Next thread to run. Chain */
#endif
	pcb_t			pcb;		/* Actual PCB */
} pthread_thread_t;
#define NULL_THREADPTR		((pthread_thread_t *) 0)

#define THREAD_CONDWAIT		0x00002
#define THREAD_SLEEPING		0x00004
#define THREAD_JOINWAIT		0x00008
#define THREAD_EXITING		0x00010
#define THREAD_EXITED		0x00020
#define THREAD_DETACHED		0x00040
#define THREAD_CANCELED		0x00080
#define THREAD_KILLED		0x00100
#define THREAD_USERSTACK	0x00200
#define THREAD_TIMEDOUT		0x00400
#define THREAD_TIMERSET		0x00800
#define THREAD_IPCSEND_WAIT	0x01000
#define THREAD_IPCRECV_WAIT	0x02000
#define THREAD_IPCREPL_WAIT	0x04000
#define THREAD_IPCWAIT_FLAG	0x07000
#define THREAD_MQREPL_WAIT	0x10000
#define THREAD_CPUIRECV_WAIT	0x20000

/*
 * When a thread gives up the CPU, tell the reschedule code why. This allows
 * for more control with regards to scheduling policy.
 */
typedef enum {
	RESCHED_PREEMPT		= 1,	/* Time based preemption yield */
	RESCHED_YIELD,			/* Involuntary yield */
	RESCHED_USERYIELD,		/* User directed yield */
	RESCHED_BLOCK,			/* Waiting for a resource (mutex) */
	RESCHED_INTERNAL,		/* Internal rescheduling call */
} resched_flags_t;

/* Default scheduling interval for SCHED_RR (in ticks) */
#define SCHED_RR_INTERVAL	1

/*
 * Currently executing thread and idlethreads.
 */
extern pthread_thread_t		*threads_curthreads[];
extern pthread_thread_t		*threads_idlethreads[];

#ifdef SMP
#define CURPTHREAD()		threads_curthreads[smp_find_cur_cpu()]
#define IDLETHREAD		threads_idlethreads[smp_find_cur_cpu()]
#else
#define CURPTHREAD()		threads_curthreads[0]
#define IDLETHREAD		threads_idlethreads[0]
#endif

/*
 * Interrupts and preemption blocks.
 */
extern int		threads_interrupt_state[];
#ifdef SMP
#define INTSTATE	(threads_interrupt_state[smp_find_cur_cpu()])
#else
#define INTSTATE	(threads_interrupt_state[0])
#endif

#define SPLHIGH		0x01
#define SPLPREEMPT	0x02
#define	INTSBLOCKED	(INTSTATE & SPLHIGH)
#define PREEMPTBLOCKED	(INTSTATE & SPLPREEMPT)

void	machine_intr_disable(void);
void	machine_intr_enable(void);
int	machine_intr_enabled(void);

#ifdef THREADS_DEBUG
int intscheck();
#else
#define intscheck()
/*
 * If the tree is built with unix examples enabled, the interrupt
 * enable/disable routines cannot be inlined with privledged commands.
 */
#ifndef UNIX_EXAMPLES
#include <oskit/x86/proc_reg.h>
#include <oskit/x86/eflags.h>
#define machine_intr_enable()	(sti())
#define machine_intr_disable()	(cli())
#define machine_intr_enabled()  (get_eflags() & EFL_IF)
#endif
#endif

OSKIT_INLINE int
splhigh(void)
{
	int	intstate = INTSTATE;

	intscheck();

	if (intstate & SPLHIGH)
		return intstate;

	machine_intr_disable();
	INTSTATE |= SPLHIGH;

	return intstate;
}

OSKIT_INLINE int
spllow(void)
{
	int	intstate = INTSTATE;

	intscheck();
	INTSTATE &= ~SPLHIGH;
	machine_intr_enable();

	return intstate;
}

OSKIT_INLINE int
splpreempt(void)
{
	int	intstate = INTSTATE;

	intscheck();
	INTSTATE |= SPLPREEMPT;

	return intstate;
}

OSKIT_INLINE int
splx(int x)
{
	int	intstate = INTSTATE;
	
	intscheck();

	if (x == intstate)
		return intstate;

	if (x & SPLHIGH) {
		machine_intr_disable();
		INTSTATE = x;
	}
	else {
		INTSTATE = x;
		machine_intr_enable();
	}

	return intstate;
}

OSKIT_INLINE int
intsenabled(void)
{
	return ((INTSTATE & SPLHIGH) == 0);
}

/*
 * Scheduler definitions. The scheduler module needs to define these
 * symbols.
 */
void			pthread_init_scheduler(void);
int			pthread_sched_reschedule(resched_flags_t reason,
				pthread_lock_t *lock);
int			pthread_sched_setrunnable(pthread_thread_t *pthread);
void			pthread_sched_deschedule(pthread_thread_t *pthread);
int			pthread_sched_change_state(pthread_thread_t *pthread,
				int newprio, int policy);
void			pthread_sched_handoff(int waitstate,
				pthread_thread_t *pnext);
int			pthread_sched_thread_wakeup(int waitstate,
				pthread_thread_t *pthread);

/*
 * The maximum number of allowed threads.
 */
#define THREADS_MAX_THREAD	512

/*
 * The array of thread structure pointers, indexed by TID.
 */
extern pthread_thread_t		*threads_tidtothread[];
#define tidtothread(tid)	(threads_tidtothread[(tid)])

/*
 * Callouts.
 */
extern void		       *(*threads_allocator)(size_t); 
extern void			(*threads_deallocator)(void *);

/*
 * Key table stuff.
 */
struct keytable {
	int		inuse;
	void		(*destructor)(void *);
};

extern pthread_lock_t		threads_key_lock;
extern struct keytable		threads_key_table[];

OSKIT_INLINE int
validkey(int key)
{
	return (key >= 0 && key < PTHREAD_KEYS_MAX &&
		threads_key_table[key].inuse);
}

/*
 * The switch routine is dependent on user/real mode operation, so its a
 * function pointer.
 */
extern void	  (*thread_switch)(pthread_thread_t *pnext,
				   pthread_lock_t *l, pthread_thread_t **cur);

/*
 * Internal Prototypes.
 */
pthread_thread_t *pthread_init_mainthread(pthread_thread_t *pthread);
void		  thread_setup(pthread_thread_t *pthread,
			void (*function)(void *), void *argument);
void		  thread_destroy(pthread_thread_t *pthread);
void	         *pthread_alloc_memory(size_t size);
int		  pthread_mprotect(oskit_addr_t addr, size_t length,
			int writeable);
void		  pthread_dealloc_memory(void *pmem);
void		  pthread_idle_function(void *argument);
pthread_thread_t *pthread_create_internal(void (*function)(void *),
			void *argument, const pthread_attr_t *pattr);
void		  pthread_destroy_internal(pthread_thread_t *pthread);
void		  pthread_block_onQ(queue_head_t *q, pthread_lock_t *plock);
pthread_thread_t *pthread_dequeue_fromQ(queue_head_t *q);
pthread_thread_t *pthread_remove_fromQ(queue_head_t *q, pthread_thread_t *pth);
void		  pthread_resume_allQ(queue_head_t *q);
void		  pthread_blockints(void);
void		  pthread_unblockints(void);
int		  pthread_wakeup_locked(pthread_thread_t *pthread);
int		  pthread_wakeup_unlocked(pthread_thread_t *pthread);
pthread_thread_t *pthread_current(void);
int		  pthread_setprio_internal(pthread_thread_t *pthread, int pri);
void		  pthread_preempt(void);
void		  pthread_yield(void);
int		  threads_stack_back_trace(int tid, int max_st_levels);
void		  thread_getstate(pthread_thread_t *pth, pthread_state_t *pst);
void		  pthread_call_key_destructors(void);
void		  pthread_call_cleanup_handlers(void);
void		  osenv_process_release(void);
int		  osenv_process_locked(void);
void		  dump_all_threads(void);
void		  pthread_mutex_panic(pthread_mutex_t *m, char *t, char *msg);
void		  pthread_delay(void);

int		  pthread_cond_wait_safe(pthread_cond_t *c,
			pthread_mutex_t *m);
void		  pthread_cond_wait_other(pthread_cond_t *c,
			pthread_thread_t *pthread);
pthread_thread_t *pthread_cond_wakeup(pthread_cond_t *c);
void		  pthread_cond_wakeup_other(pthread_cond_t *c,
			pthread_thread_t *pthread);

void		  pthread_exit_locked(void *status);
void		  pthread_reap_threads(void);

int		  pthread_init_comlock(void);
void		  pthread_init_attributes(void);
void		  pthread_init_guard(void);
void		  pthread_init_libc_locks(void);
void		  pthread_init_osenv_sleep(void);
void		  pthread_init_osenv_intr(void);
void		  pthread_init_keytable(void);
void		  pthread_init_process_lock(void);
void		  pthread_init_fs_sleep(void);
void		  pthread_init_exit(void);
void		  pthread_init_signals(void);
void		  pthread_ipc_cancel(pthread_thread_t *pthread);
void		  oskit_init_libc(void);
void		  thread_machdep_init(void);
oskit_error_t	  pthread_register_interface(void);
void		  oskit_deliver_pending_signals(void);

#ifdef PRI_INHERIT
void		  pthread_priority_inherit(pthread_thread_t *pthread);
void		  pthread_priority_uninherit(pthread_thread_t *pthread);
void		  pthread_priority_decreasing_recompute(pthread_thread_t *pth);
void		  pthread_priority_increasing_recompute(pthread_thread_t *pth);
void		  pthread_priority_kill(pthread_thread_t *pthread);
#endif

#ifndef PAGE_SIZE
#define PAGE_SIZE	0x1000
#endif

#ifdef SMP
#define MAXCPUS		8
#define THISCPU		(smp_find_cur_cpu())
#else
#define MAXCPUS		1
#define THISCPU		0
#endif

#ifdef THREADS_DEBUG
#undef  STACKGUARD
#define STACKGUARD

/*
 * Deadlock detection.
 */
extern int		threads_sleepers;
extern pthread_lock_t	threads_sleepers_lock;
#endif

extern int		threads_debug;
extern pthread_thread_t threads_mainthread;

/*
 * Soft interrupt support.
 *
 * There are two situations in which a softint handler will be requested.
 * The first is from the time-based preemption code. The second is
 * when a wakeup operation (resume or timed condtion timeout) occurs, 
 * and the priorities have changed, requiring a reschedule operation.
 */
extern int			threads_switch_mode;

#define SOFTINT_TIMEOUT		0x01
#define SOFTINT_ASYNCREQ	0x02

/*
 * Seems kinda non-portable, don't ya think?
 */
#include <oskit/x86/pc/base_irq.h>

OSKIT_INLINE void
softint_request(int type)
{
	threads_switch_mode |= type;
	base_irq_softint_request();
}

/*
 * Check for directed yield.
 */
OSKIT_INLINE void
check_yield()
{
	if (CURPTHREAD()->preempt && !(INTSBLOCKED || PREEMPTBLOCKED))
		pthread_yield();
}

OSKIT_INLINE void
queue_check(queue_head_t *queue, pthread_thread_t *pthread)
{
	pthread_thread_t	*pnext;

	queue_iterate(queue, pnext, pthread_thread_t *, chain) {
		if (pnext == pthread)
			panic("queue_check");
	}
}

#endif /* ASSEMBLER */

/*
 * Define assembly language constants.
 */
#define THREAD_NEXT		0x0
#define THREAD_PREV		0x4
#define THREAD_PPCB		0x8

#endif /* _THREADS_INT_H_ */
