/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Posix signals, or something approximating them. The goal is to get
 * the obvious stuff nearly correct, but not go overboard.
 */

#include <threads/pthread_internal.h>
#include <strings.h>

#define IN_AN_INTERRUPT()	(base_irq_nest & ~0x80)

/*
 * The sigaction data is kept in the threads library.
 */
static struct sigaction sigactions[NSIG];

/*
 * Lock to protect the sigaction structure.
 */
static pthread_lock_t	siglock = PTHREAD_LOCK_INITIALIZER;

void	really_deliver_signal(int sig, siginfo_t *code,struct sigcontext *scp);
void    oskit_deliver_sync_signal(int sig);

int
pthread_sigmask(int how, const sigset_t *set, sigset_t *oset)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	int			p, err = 0;

	p = splhigh();
	pthread_lock(&pthread->lock);

	if (oset)
		*oset = pthread->sigmask;

	if (set) {
		switch (how) {
		case SIG_BLOCK:
			pthread->sigmask |= *set;
			break;
		case SIG_UNBLOCK:
			pthread->sigmask &= ~*set;
			break;
		case SIG_SETMASK:
			pthread->sigmask = *set;
			break;
		default:
			err = EINVAL;
		}
	}

	/*
	 * Look for pending signals that are now unblocked, and deliver.
	 */
	while (pthread->sigpending & ~pthread->sigmask) {
		int sig = ffs(pthread->sigpending & ~pthread->sigmask);

		/* Call at splhigh and thread locked */
		oskit_deliver_sync_signal(sig);
	}

	pthread_unlock(&pthread->lock);
	splx(p);
	return err;
}

int
pthread_kill(pthread_t tid, int signo)
{
	pthread_thread_t	*pthread;
	int			p;

	if (signo <= 0 || signo >= NSIG) {
		printf("pthread_kill: Illegal signo %d\n", signo);
		return EINVAL;
	}

	if (tid < 0 || tid >= THREADS_MAX_THREAD ||
	    ((pthread = tidtothread(tid)) == NULL_THREADPTR))
		return EINVAL;

	p = splhigh();
	pthread_lock(&siglock);

	/*
	 * Look at the process sigactions. If the "process" is ignoring
	 * the signal, then the signal is not placed in the pending list.
	 */
	if (sigactions[signo].sa_handler == SIG_IGN) {
		pthread_unlock(&siglock);
		splx(p);
		return 0;
	}
	pthread_unlock(&siglock);
	pthread_lock(&pthread->lock);

	/*
	 * If the thread is on its way out, forget about it.
	 */
	if (pthread->flags & (THREAD_EXITED|THREAD_EXITING|THREAD_CANCELED)) {
		pthread_unlock(&pthread->lock);
		splx(p);
		return 0;
	}

	/*
	 * Add the signal to list of pending signals for the target thread.
	 */
	sigaddset(&pthread->sigpending, signo);

	/*
	 * If the signal is currently blocked, then do nothing else.
	 * It will be noticed when the signal is unblocked.
	 */
	if (sigismember(&pthread->sigmask, signo)) {
		pthread_unlock(&pthread->lock);
		splx(p);
		return 0;
	}

	/*
	 * Is the current thread sending itself a signal? This is okay, as
	 * long as its not from within an interrupt handler, say, from an
	 * alarm expiration. The signal can be delivered immediately.
	 *
	 * If this *is* from within an interrupt, must wait until later
	 * since delivering it now is really the wrong way to go. Lets
	 * schedule a soft interrupt, and get the thread switched out.
	 * The signal will be delivered when it comes back in.
	 */
	if (pthread == CURPTHREAD()) {
		if (IN_AN_INTERRUPT()) {
			softint_request(SOFTINT_ASYNCREQ);
			pthread_unlock(&pthread->lock);
			splx(p);
		}
		else {
			/* Call at splhigh and thread locked */
			oskit_deliver_sync_signal(signo);
			pthread_unlock(&pthread->lock);
			splx(p);
		}
		return 0;
	}

	/*
	 * Now look at what the thread is doing, and restart it.
	 * The signal will be noticed when the thread wakes up and
	 * returns from its block.
	 *
	 * XXX This is very similar to what pthread_cancel does ...
	 */

	/*
	 * IPC Wait.
	 */
	if (pthread->flags & THREAD_IPCWAIT_FLAG) {
		pthread_ipc_cancel(pthread);
		goto done;
	}
	
	/*
	 * If the thread is in an osenv_sleep(), issue a wakeup.
	 * The thread will be allowed to return through the sleep, to
	 * be caught sometime later. This allows driver state to be
	 * cleaned up before the thread is actually killed off.
	 */
	if (pthread->sleeprec) {
		pthread_unlock(&pthread->lock);
		osenv_wakeup(pthread->sleeprec, OSENV_SLEEP_CANCELED);
		goto done;
	}

	/*
	 * If the thread is THREAD_SLEEPING, then restart it. The death 
	 * will be noticed before the thread is allowed to return from 
	 * the call.
	 */
	if (pthread->flags & THREAD_SLEEPING) {
		pthread_wakeup_locked(pthread);
		goto done;
	}

	/*
	 * If the thread is THREAD_CONDWAIT, then restart it. The wrinkle
	 * is a race condition between the time the thread is taken off
	 * the condition queue and the time the thread state is changed.
	 */
	if (pthread->flags & THREAD_CONDWAIT) {
		pthread_cond_t	*c = pthread->waitcond;

		pthread_lock(&(c->lock));

		/*
		 * The thread was still on the Q, so its safe to change
		 * its state to reflect that it is not longer waiting
		 * on the condition. 
		 *
		 * If the thread was not on the Q, we caught the race,
		 * and do not have to do anything.
		 */
		if (pthread_remove_fromQ(&(c->waiters), pthread)) {
			pthread->flags   &= ~THREAD_CONDWAIT;
			pthread->waitcond = 0;
			pthread_unlock(&(c->lock));
			pthread_unlock(&pthread->lock);
			pthread_sched_setrunnable(pthread);
			goto done;
		}

		pthread_unlock(&pthread->lock);
		pthread_unlock(&(c->lock));
		goto done;
	}

	/*
	 * Must be running on another CPU. Must wait for it to be noticed.
	 */
	pthread_unlock(&pthread->lock);

 done:
	splx(p);
	return 0;
}

/*
 * The rest of the Posix signal code. This stuff replaces the corresponding
 * code in the Posix library, since I don't see how it can be done unless
 * its all in one place.
 */

/*
 * sigaction
 */
int
sigaction(int sig, const struct sigaction *act, struct sigaction *oact)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	int			s;
	
	if (sig < 0 || sig >= NSIG)
		return errno = EINVAL, -1;

	s = splhigh();
	pthread_lock(&siglock);

	if (oact)
		*oact = sigactions[sig];
	if (act)
		sigactions[sig] = *act;

	pthread_unlock(&siglock);

	/*
	 * If the action for this signal is being set to SIG_IGN,
	 * and that signal is pending, then clear it.
	 */
	if (act && act->sa_handler == SIG_IGN) {
		pthread_lock(&pthread->lock);
		sigdelset(&pthread->sigpending, sig);
		pthread_unlock(&pthread->lock);
	}

	splx(s);
	return 0;
}

/*
 * sigprocmask. In a multithreaded program, this is just pthread_sigmask
 */
int
sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
	return pthread_sigmask(how, set, oset);
}

/*
 * raise. In a multithreaded program, raise is pthread_kill on itself.
 */
int
raise(int sig)
{
	return pthread_kill(pthread_self(), sig);
}

/*
 * kill. This is the process equivalent, which in the multithreaded
 * case means find some thread that can take the signal, and send it
 * the pthread_kill. We deviate from the rules here, since there is
 * no process global sigpending and no sigwait implementation. If all
 * threads are blocking the signal, just pick one to send it to.
 */
int
kill(pid_t pid, int sig)
{
	pthread_thread_t	*pthread;
	int			i, p;

	p = splhigh();
	for (i = 1; i < THREADS_MAX_THREAD; i++) {
		if ((pthread = threads_tidtothread[i]) != NULL) {
			pthread_lock(&pthread->lock);

			/*
			 * If the thread is on its way out or joining,
			 * forget about it.
			 */
			if (pthread->flags &
			    (THREAD_EXITED|THREAD_EXITING|
			     THREAD_CANCELED|THREAD_JOINWAIT)) {
				    pthread_unlock(&pthread->lock);
				    continue;
			}

			/*
			 * If the signal is currently unblocked, deliver
			 * to this thread. Don't worry about the race.
			 */
			if (! sigismember(&pthread->sigmask, sig)) {
				pthread_unlock(&pthread->lock);
				splx(p);
				return pthread_kill(i, sig);
			}
			
			pthread_unlock(&pthread->lock);
		}
	}
	/*
	 * No thread was ready to take the signal. Just post it for the
	 * first available one.
	 */
	for (i = 1; i < THREADS_MAX_THREAD; i++) {
		if ((pthread = threads_tidtothread[i]) != NULL) {
			pthread_lock(&pthread->lock);

			/*
			 * If the thread is on its way out or joining,
			 * forget about it.
			 */
			if (pthread->flags &
			    (THREAD_EXITED|THREAD_EXITING|
			     THREAD_CANCELED|THREAD_JOINWAIT)) {
				    pthread_unlock(&pthread->lock);
				    continue;
			}

			pthread_unlock(&pthread->lock);
			splx(p);
			return pthread_kill(i, sig);
		}
	}
	splx(p);
	panic("kill: Could not find a thread to deliver to!");
	return 0;
}

/*
 * Internal stuff.
 */

/*
 * Deliver a signal generated from a trap. This is the upcall from the
 * machine dependent code that created the sigcontext structure from
 * the trap state.
 */
void
oskit_libc_sendsig(int sig, int code, struct sigcontext *scp)
{
	siginfo_t	siginfo;

	siginfo.si_signo           = sig;
	siginfo.si_code            = SI_EXCEP;
	siginfo.si_value.sival_int = code;

	really_deliver_signal(sig, &siginfo, scp);
}

/*
 * Deliver a synchronous signal.
 */
void
oskit_deliver_sync_signal(int sig)
{
	struct sigcontext	sc;
	siginfo_t		siginfo;

	/* create a stub sigcontext_t. */
	bzero(&sc, sizeof(sc));

	siginfo.si_signo           = sig;
	siginfo.si_code            = SI_USER;
	siginfo.si_value.sival_int = 0;

	really_deliver_signal(sig, &siginfo, &sc);
}

/*
 * Deliver any pending signals. Called out of the context switch code
 * when a thread switches in, and there are pending signals.
 *
 * Interrupts are blocked and the thread is locked.
 */
void
oskit_deliver_pending_signals(void)
{
	pthread_thread_t	*pthread = CURPTHREAD();

	while (pthread->sigpending & ~pthread->sigmask) {
		int sig = ffs(pthread->sigpending & ~pthread->sigmask);

		/* Call at splhigh and thread locked */
		oskit_deliver_sync_signal(sig);
	}
}

/*
 * Actually deliver the signal to the thread. At this point the signal
 * is going to be delivered, so it no longer matters if it is blocked.
 */
void
really_deliver_signal(int sig, siginfo_t *info, struct sigcontext *scp)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	sigset_t		sigmask, oldmask;
	struct sigaction	act;

	pthread_lock(&siglock);
	act = sigactions[sig];
	pthread_unlock(&siglock);

	/*
	 * Ignored?
	 */
	if (act.sa_handler == SIG_IGN || act.sa_handler == SIG_ERR)
		return;

	if (act.sa_handler == SIG_DFL) {
		/* Default action for all signals is termination */
		panic("Sendsig: Signal %d caught but no handler", sig);
	}

	/*
	 * Set the signal mask for calling the handler.
	 */
	oldmask = sigmask = pthread->sigmask;
	sigaddset(&sigmask, sig);
	sigmask |= act.sa_mask;
	sigdelset(&pthread->sigpending, sig);
	pthread->sigmask = sigmask;
	pthread_unlock(&pthread->lock);
	spllow();

	/*
	 * and call the handler ...
	 */
	if (act.sa_flags & SA_SIGINFO)
		act.sa_sigaction(sig, info, (void *) scp);
	else
		((void (*)(int, int, struct sigcontext *))act.sa_handler)
			(sig, info->si_value.sival_int, scp);

	splhigh();
	pthread_lock(&pthread->lock);
	pthread->sigmask = oldmask;
	
	/*
	 * and return with thread locked at splhigh.
	 */
}

void
pthread_init_signals()
{
	int	i;

	/* Initialize the default signal actions. */
	for (i = 0; i < NSIG; i++)
		sigactions[i].sa_handler = SIG_DFL;

	libc_sendsig_init();
}

void
signals_init(void)
{
	/* Called by libc. Initialization was already done above. */ 
}

